"""Utility functions to help with testing Language Server features."""
import logging
import pathlib
from typing import List
from typing import Optional
from typing import Set

from pygls.lsp.types import Position
from pygls.workspace import Document
from sphinx import __version__ as __sphinx_version__

from esbonio.lsp import LanguageFeature

logger = logging.getLogger(__name__)


def sphinx_version(eq: Optional[int] = None) -> bool:
    """Helper function for determining which version of Sphinx we are
    testing with.

    Currently this only cares about the major version number.

    Parameters
    ----------
    eq:
       When set returns ``True`` if the Sphinx version exactly matches
       what's given.
    """

    major, _, _ = [int(v) for v in __sphinx_version__.split(".")]

    if eq and major == eq:
        return True

    return False


def directive_argument_patterns(name: str, partial: str = "") -> List[str]:
    """Return a number of example directive argument patterns.

    These correspond to test cases where directive argument suggestions should be
    generated.

    Parameters
    ----------
    name:
       The name of the directive to generate suggestions for.
    partial:
       The partial argument that the user has already entered.
    """
    return [s.format(name, partial) for s in [".. {}:: {}", "   .. {}:: {}"]]


def role_patterns(partial: str = "") -> List[str]:
    """Return a number of example role patterns.

    These correspond to when role suggestions should be generated.

    Parameters
    ----------
    partial:
       The partial role name that the user has already entered
    """
    return [
        s.format(partial)
        for s in [
            "{}",
            "({}",
            "   {}",
            "   ({}",
            "some text {}",
            "some text ({}",
            "   some text {}",
            "   some text ({}",
        ]
    ]


def role_target_patterns(name: str, partial: str = "") -> List[str]:
    """Return a number of example role target patterns.

    These correspond to test cases where role target suggestions should be generated.

    Parameters
    ----------
    name:
       The name of the role to generate suggestions for.
    partial:
       The partial target that the user as already entered.
    """
    return [
        s.format(name, partial)
        for s in [
            ":{}:`{}",
            "(:{}:`{}",
            ":{}:`More Info <{}",
            "(:{}:`More Info <{}",
            "   :{}:`{}",
            "   (:{}:`{}",
            "   :{}:`Some Label <{}",
            "   (:{}:`Some Label <{}",
        ]
    ]


def intersphinx_target_patterns(name: str, project: str) -> List[str]:
    """Return a number of example intersphinx target patterns.

    These correspond to cases where target completions may be generated

    Parameters
    ----------
    name: str
       The name of the role to generate examples for
    project: str
       The name of the project to generate examples for
    """
    return [
        s.format(name, project)
        for s in [
            ":{}:`{}:",
            "(:{}:`{}:",
            ":{}:`More Info <{}:",
            "(:{}:`More Info <{}:",
            "   :{}:`{}:",
            "   (:{}:`{}:",
            "   :{}:`Some Label <{}:",
            "   (:{}:`Some Label <{}:",
        ]
    ]


def completion_test(
    feature: LanguageFeature,
    text: str,
    *,
    filepath: str = "index.rst",
    expected: Optional[Set[str]] = None,
    unexpected: Optional[Set[str]] = None,
):
    """Check to see if a feature provides the correct completion suggestions.

    .. admonition:: Assumptions

       - The ``feature`` has access to a valid Sphinx application via ``rst.app``
       - If ``feature`` requires initialisation, it has already been done.

    .. admonition:: Limitations

       - Only checking ``CompletionItem`` labels is supported, if you want to check
         other aspects of the results, write a dedicated test method.

    This function takes the given ``feature`` and calls it in the same manner as the
    real language server so that it can simulate real usage without being a full blown
    integration test.

    This requires ``suggest_triggers`` to be set and it to have a working ``suggest``
    method.

    Completions will be asked for with the cursor's position to be at the end of the
    inserted ``text`` in a blank document by default. If your test case requires
    additional context this can be included in ``text`` delimited by a ``\\f`` character.

    For example to pass text representing the following scenario (``^`` represents the
    user's cursor)::

       .. image:: filename.png
          :align: center
          :
           ^

    The ``text`` parameter should be set to
    ``.. image:: filename.png\\n   :align: center\\n\\f   :``. It's important to note
    that newlines **cannot** come after the ``\\f`` character.

    If you want to test the case where no completions should be suggested, set
    ``expected`` to ``None``.

    By default completion suggestions will be asked for the main ``index.rst`` file at
    the root of the docs project. If you want them to be asked for a file with a
    different path (like for filepath completion tests) this can be overridden with the
    ``filepath`` argument.

    Parameters
    ----------
    feature:
       An instance of the language service feature to test.
    text:
       The text to offer completion suggestions for.
    filepath:
       The path of the file that completion suggestions are being asked for. Relative
       to the Sphinx application's srcdir.
    expected:
       The set of completion item labels you expect to see in the output.
    unexpected:
       The set of completion item labels you do *not* expect to see in the output.
    """

    if "\f" in text:
        contents, text = text.split("\f")
    else:
        contents = ""

    logger.debug("Context text:    '%s'", contents)
    logger.debug("Insertion text: '%s'", text)
    assert "\n" not in text, "Insertion text cannot contain newlines"

    filepath = pathlib.Path(feature.rst.app.srcdir) / filepath
    document = Document(f"file://{filepath}", contents)
    position = Position(line=len(document.lines), character=len(text) - 1)

    results = []
    for trigger in feature.suggest_triggers:
        match = trigger.match(text)
        logger.debug("Match: %s", match)

        if match:
            results += feature.suggest(match, document, position)

    items = {item.label for item in results}
    unexpected = unexpected or set()

    logger.debug("Results:    %s", items)
    logger.debug("Expected:   %s", expected)
    logger.debug("Unexpected: %s", unexpected)

    if expected is None:
        assert len(items) == 0
    else:
        assert expected == items & expected
        assert set() == items & unexpected



"""Intersphinx support."""
import re
from typing import List
from typing import Optional

from pygls.lsp.types import CompletionItem
from pygls.lsp.types import CompletionItemKind
from pygls.lsp.types import Position
from pygls.workspace import Document

import esbonio.lsp as lsp
from esbonio.lsp.roles import COMPLETION_TARGETS
from esbonio.lsp.roles import DEFAULT_TARGET
from esbonio.lsp.roles import PARTIAL_ALIASED_TARGET
from esbonio.lsp.roles import PARTIAL_PLAIN_TARGET
from esbonio.lsp.sphinx import get_domains


PARTIAL_INTER_PLAIN_TARGET = re.compile(
    r"""
    (^|.*[^\w:])             # roles cannot be preceeded by certain chars
    (?P<role>:               # roles start with the ':' character
    (?!:)                    # make sure the next character is not ':'
    (?P<domain>[\w]+:)?      # there may be a domain namespace
    (?P<name>[\w-]*)         # followed by the role name
    :)                       # the role name ends with a ':'
    `                        # the target begins with a '`'
    (?P<project>[^<:`]*)     # match "plain link" targets
    :                        # projects end with a ':'
    $
    """,
    re.MULTILINE | re.VERBOSE,
)
"""A regular expression that matches a partial "plain" intersphinx target.

For example::

   :ref:`python:som

Used when generating auto complete suggestions.
"""

PARTIAL_INTER_ALIASED_TARGET = re.compile(
    r"""
    (^|.*[^\w:])          # roles cannot be preceeded by alpha chars
    (?P<role>:            # roles start with the ':' character
    (?!:)                 # make sure the next character is not ':'
    (?P<domain>[\w]+:)?   # there may be a domain namespace
    (?P<name>[\w-]*)      # followed by the role name
    :)                    # the role name ends with a ':'
    `                     # the target begins with a '`'`
    .*<                   # the actual target name starts after a '<'
    (?P<project>[^`:]*)   # match "aliased" targets
    :                     # projects end with a ':'
    $
    """,
    re.MULTILINE | re.VERBOSE,
)
"""A regular expression that matches a partial "aliased" intersphinx target.

For example::

   :ref:`More Info<python:som

Used when generating auto complete suggestions.
"""


class InterSphinx(lsp.LanguageFeature):
    """Intersphinx support for the language server."""

    def initialize(self, options: lsp.InitializationOptions):

        self.targets = {}
        self.target_types = {}
        self.projects = {}

        if self.rst.app and hasattr(self.rst.app.env, "intersphinx_named_inventory"):
            self.discover_inventories()

    def discover_inventories(self):
        """Look up intersphinx inventories to offer as autocomplete suggestions.

        *This method only needs to be called once per application instance.*

        Assuming that the application has the ``intersphinx`` extension enabled and
        configured, this will look for the ``intersphinx_named_inventory`` attribute
        on the Sphinx application's build environment and index each of the availble
        projects.

        This will also loop through each domain object and construct the
        ``target_types`` dictionary - very similar to
        :meth:`~esbonio.lsp.roles.Roles.discover_roles` on the ``Roles`` feature, except
        that intersphinx always uses a domain's namespace even if ``primary_domain`` is
        set in the app's config.

        Finally, as target completions for other projects aren't going to change over
        the course of an editing session, it's also safe to index them once on init.
        """

        inv = self.rst.app.env.intersphinx_named_inventory
        self.projects = {v: self.project_to_completion_item(v) for v in inv.keys()}

        for _, domain in get_domains(self.rst.app):

            # Intersphinx entries are always namespaced, regardless of the
            # `primary_domain` setting.
            fmt = "{domain.name}:{name}"

            # Build a map we can use to lookup target completions
            for name, item_type in domain.object_types.items():
                for role in item_type.roles:
                    key = fmt.format(name=role, domain=domain)
                    target_types = self.target_types.get(key, None)

                    if target_types is None:
                        target_types = []

                    target_types.append(fmt.format(name=name, domain=domain))
                    self.target_types[key] = target_types

            # Index each of the completion targets.
            for project, types in inv.items():
                project_map = {}

                for target_type, targets in types.items():
                    target_type_map = {}

                    for label, target in targets.items():
                        target_type_map[label] = self.target_to_completion_item(
                            label, target, target_type
                        )

                    project_map[target_type] = target_type_map

                self.targets[project] = project_map

        self.logger.info("Discovered %s intersphinx projects", len(self.projects))
        self.logger.debug("Projects: %s", self.projects.keys())

        self.logger.info("Discovered %s target types", len(self.target_types))
        self.logger.debug("Target types: %s", self.target_types)

    suggest_triggers = [
        PARTIAL_PLAIN_TARGET,
        PARTIAL_ALIASED_TARGET,
        PARTIAL_INTER_PLAIN_TARGET,
        PARTIAL_INTER_ALIASED_TARGET,
    ]

    def suggest(
        self, match: "re.Match", doc: Document, position: Position
    ) -> List[CompletionItem]:

        # As a unfortunate consequence of naming choices, we're in the counter intuitive
        # situation where
        #
        # - A match containing a "target" regex group should suggest project names
        # - A match containing a "project" regex group should suggest targets
        groups = match.groupdict()

        if "target" in groups:
            return self.suggest_projects(match)

        return self.suggest_targets(match)

    def suggest_projects(self, match: "re.Match") -> List[CompletionItem]:

        if self.get_target_types(match):
            self.logger.info("Suggesting projects")
            return list(self.projects.values())

        return []

    def suggest_targets(self, match: "re.Match") -> List[CompletionItem]:
        # TODO: Detect if we're in an angle bracket e.g. :ref:`More Info <python:`
        # and add the closing '>' to the completion item insert text.
        self.logger.info("Suggesting targets")

        target_types = self.get_target_types(match)
        if target_types is None:
            return []

        project = self.targets.get(match.group("project"), None)
        if project is None:
            return []

        targets = []
        for target_type in target_types:
            items = project.get(target_type, {})
            targets += items.values()

        return targets

    def get_target_types(self, match: "re.Match") -> Optional[List[str]]:
        """Returns the list of target types that are targeted by the role we're
        generating suggestions for."""

        role = match.group("name")
        domain = match.group("domain") or ""
        primary_domain = self.rst.app.config.primary_domain or ""

        self.logger.debug("Looking up target types for '%s%s'", domain, role)

        # Attempt to find the right key..
        for key in [f"{domain}{role}", f"{primary_domain}:{role}", f"std:{role}"]:
            target_types = self.target_types.get(key, None)
            self.logger.debug("Targets types for '%s': %s", key, target_types)

            if target_types is not None:
                return target_types

    def project_to_completion_item(self, project: str) -> CompletionItem:
        return CompletionItem(
            label=project, detail="intersphinx", kind=CompletionItemKind.Module
        )

    def target_to_completion_item(
        self, label: str, target, target_type: str
    ) -> CompletionItem:

        key = target_type

        if ":" in key:
            key = ":".join(key.split(":")[1:])

        completion_type = COMPLETION_TARGETS.get(key, DEFAULT_TARGET)
        source, version, _, display = target

        if display == "-":
            display = label

        if version:
            version = f" v{version}"

        detail = f"{display} - {source}{version}"

        return CompletionItem(
            label=label, kind=completion_type.kind, detail=detail, insert_text=label
        )


def setup(rst: lsp.RstLanguageServer):
    intersphinx = InterSphinx(rst)
    rst.add_feature(intersphinx)



"""This module defines a custom logging handler that publishes log messages to an LSP
client."""
import logging

from pygls.server import LanguageServer

LOG_LEVELS = {
    "debug": logging.DEBUG,
    "error": logging.ERROR,
    "info": logging.INFO,
}


class LogFilter:
    """A log filter that accepts message from any of the listed logger names."""

    def __init__(self, names):
        self.names = names

    def filter(self, record):
        return any(record.name == name for name in self.names)


class LspHandler(logging.Handler):
    """A logging handler that will send log records to an LSP client."""

    def __init__(self, server: LanguageServer):
        super().__init__()
        self.server = server

    def emit(self, record: logging.LogRecord) -> None:
        """Sends the record to the client."""

        # To avoid infinite recursions, it's simpler to just ignore all log records
        # coming from pygls...
        if "pygls" in record.name:
            return

        log = self.format(record).strip()
        self.server.show_message_log(log)



"""Role support."""
import collections
import re
from typing import List

from docutils.parsers.rst import roles
from pygls.lsp.types import CompletionItem
from pygls.lsp.types import CompletionItemKind
from pygls.lsp.types import DidSaveTextDocumentParams
from pygls.lsp.types import Position
from pygls.lsp.types import Range
from pygls.lsp.types import TextEdit
from pygls.workspace import Document

import esbonio.lsp as lsp
from esbonio.lsp.directives import DIRECTIVE
from esbonio.lsp.sphinx import get_domains


PARTIAL_ROLE = re.compile(
    r"""
    (^|.*[^\w:])          # roles cannot be preceeded by certain characters
    (?P<role>:            # roles start with the ':' character
    (?!:)                 # make sure the next character is not ':'
    (?P<domain>[\w]+:)?   # there may be a domain namespace
    (?P<name>[\w-]*))     # match the role name
    $                     # ensure pattern only matches incomplete roles
    """,
    re.MULTILINE | re.VERBOSE,
)
"""A regular expression that matches a partial role.

For example::

   :re

Used when generating auto complete suggestions.
"""


PARTIAL_PLAIN_TARGET = re.compile(
    r"""
    (^|.*[^\w:])          # roles cannot be preceeded by certain chars
    (?P<role>:            # roles start with the ':' character
    (?!:)                 # make sure the next character is not ':'
    (?P<domain>[\w]+:)?   # there may be a domain namespace
    (?P<name>[\w-]*)      # followed by the role name
    :)                    # the role name ends with a ':'
    `                     # the target begins with a '`'
    (?P<target>[^<:`]*)   # match "plain link" targets
    $
    """,
    re.MULTILINE | re.VERBOSE,
)
"""A regular expression that matches a partial "plain" role target.

For example::

   :ref:`som

Used when generating auto complete suggestions.
"""

PARTIAL_ALIASED_TARGET = re.compile(
    r"""
    (^|.*[^\w:])          # roles cannot be preceeded by certain chars
    (?P<role>:            # roles start with the ':' character
    (?!:)                 # make sure the next character is not ':'
    (?P<domain>[\w]+:)?   # there may be a domain namespace
    (?P<name>[\w-]*)      # followed by the role name
    :)                    # the role name ends with a ':'
    `                     # the target begins with a '`'`
    .*<                   # the actual target name starts after a '<'
    (?P<target>[^`:]*)    # match "aliased" targets
    $
    """,
    re.MULTILINE | re.VERBOSE,
)
"""A regular expression that matches an "aliased" role target.

For example::

   :ref:`More info <som

Used when generating auto complete suggestions.
"""


CompletionTarget = collections.namedtuple("CompletionTarget", "kind,insert_fmt")

DEFAULT_TARGET = CompletionTarget(CompletionItemKind.Reference, "{name}")
COMPLETION_TARGETS = {
    "attribute": CompletionTarget(CompletionItemKind.Field, "{name}"),
    "doc": CompletionTarget(CompletionItemKind.File, "/{name}"),
    "class": CompletionTarget(CompletionItemKind.Class, "{name}"),
    "envvar": CompletionTarget(CompletionItemKind.Variable, "{name}"),
    "function": CompletionTarget(CompletionItemKind.Function, "{name}"),
    "method": CompletionTarget(CompletionItemKind.Method, "{name}"),
    "module": CompletionTarget(CompletionItemKind.Module, "{name}"),
    "term": CompletionTarget(CompletionItemKind.Text, "{name}"),
}


class Roles(lsp.LanguageFeature):
    """Role support for the language server."""

    def initialize(self, options: lsp.InitializationOptions):
        self.discover_roles()
        self.discover_targets()

    def save(self, params: DidSaveTextDocumentParams):
        self.discover_targets()

    def discover_roles(self):
        """Look up for valid role defintions to to offer as autocomplete suggestions.

        *This method only needs to be called once per application instance.*

        This will look for all the roles registered with docutils as well as the
        roles that are stored on a Sphinx domain object.

        Additionally, while we are looping through the domain objects, we construct
        the ``target_types`` dictionary. This is used when providing role target
        completions by giving the list of object types the current role is able to
        link with.

        For example, consider the :rst:role:`sphinx:py:func` and
        :rst:role:`sphinx:py:class` roles from the Python domain. As the ``func`` role
        links to Python functions and the ``class`` role links to Python classes *and*
        exceptions we would end up with

        .. code-block:: python

           {
             "func": ["py:function"],
             "class": ["py:class", "py:exception"]
           }

        """

        # Find roles that have been registered directly with docutils
        self.target_types = {}
        found_roles = {**roles._roles, **roles._role_registry}

        # Find roles under Sphinx domains
        for prefix, domain in get_domains(self.rst.app):
            fmt = "{prefix}:{name}" if prefix else "{name}"

            for name, role in domain.roles.items():
                key = fmt.format(name=name, prefix=prefix)
                found_roles[key] = role

            # Also build a map we can use when looking up target completions.
            for name, item_type in domain.object_types.items():
                for role in item_type.roles:
                    key = fmt.format(name=role, prefix=prefix)
                    target_types = self.target_types.get(key, None)

                    if target_types is None:
                        target_types = []

                    target_types.append(fmt.format(name=name, prefix=prefix))
                    self.target_types[key] = target_types

        self.roles = {
            k: v for k, v in found_roles.items() if v != roles.unimplemented_role
        }

        self.logger.info("Discovered %s roles", len(self.roles))
        self.logger.debug("Roles: %s", list(self.roles.keys()))

        self.logger.info("Discovered %s target types", len(self.target_types))
        self.logger.debug("Target types: %s", self.target_types)

    def discover_targets(self):
        """Look up all the targets we can offer as autocomplete suggestions.

        *This method needs to be called each time a document has been saved.*
        """
        self.target_objects = {}

        for prefix, domain in get_domains(self.rst.app):
            fmt = "{prefix}:{name}" if prefix else "{name}"

            for (name, display_name, obj_type, _, _, _) in domain.get_objects():
                key = fmt.format(name=obj_type, prefix=prefix)
                items = self.target_objects.get(key, None)

                if items is None:
                    items = []
                    self.target_objects[key] = items

                items.append(
                    self.target_object_to_completion_item(name, display_name, obj_type)
                )

    suggest_triggers = [PARTIAL_ROLE, PARTIAL_PLAIN_TARGET, PARTIAL_ALIASED_TARGET]

    def suggest(
        self, match: "re.Match", doc: Document, position: Position
    ) -> List[CompletionItem]:
        indent = match.group(1)

        if "target" in match.groupdict():
            return self.suggest_targets(match, position)

        # If there's no indent, then this can only be a role defn
        if indent == "":
            return self.suggest_roles(match, position)

        # Otherwise, search backwards until we find a blank line or an unindent
        # so that we can determine the appropriate context.
        linum = position.line - 1

        try:
            line = doc.lines[linum]
        except IndexError:
            return self.suggest_roles(match, position)

        while line.startswith(indent):
            linum -= 1
            line = doc.lines[linum]

        # Unless we are within a directive's options block, we should offer role
        # suggestions
        if DIRECTIVE.match(line):
            return []

        return self.suggest_roles(match, position)

    def suggest_roles(
        self, match: "re.Match", position: Position
    ) -> List[CompletionItem]:
        self.logger.info("Suggesting roles")

        domain = match.groupdict()["domain"] or ""
        items = []

        for name, role in self.roles.items():

            if not name.startswith(domain):
                continue

            item = self.role_to_completion_item(name, role, match, position)
            items.append(item)

        return items

    def suggest_targets(
        self, match: "re.Match", position: Position
    ) -> List[CompletionItem]:

        self.logger.info("Suggesting targets")

        groups = match.groupdict()
        domain = groups["domain"] or ""
        key = f"{domain}{groups['name']}"

        object_types = self.target_types.get(key, None)

        self.logger.debug("Getting suggestions for '%s'", key)
        self.logger.debug("Role targets object types: %s", object_types)

        if object_types is None:
            return []

        targets = []
        for type_ in object_types:
            targets += self.target_objects.get(type_, [])

        return targets

    def role_to_completion_item(
        self, name: str, role, match: "re.Match", position: Position
    ) -> CompletionItem:
        """Convert an rst role to its CompletionItem representation.

        With domain support it's necessary to compute the CompletionItem representation
        specifically for each completion site. See
        :meth:`~esbonio.lsp.directives.Directives.directive_to_completion_item` for
        more historical information.

        For some reason, even though these completion items are constructed in the same
        manner as the ones for directives using them in VSCode does not feel as nice....

        Parameters
        ----------
        name:
           The name of the role as a user would type into an reStructuredText document.
        role:
           The implementation of the role.
        match:
           The regular expression match object that represents the line we are providing
           the autocomplete suggestions for.
        position:
           The position in the source code where the autocompletion request was sent
           from.
        """

        groups = match.groupdict()

        line = position.line
        start = position.character - len(groups["role"])
        end = position.character

        insert_text = f":{name}:"

        item = CompletionItem(
            label=name,
            kind=CompletionItemKind.Function,
            filter_text=insert_text,
            detail="role",
            text_edit=TextEdit(
                range=Range(
                    start=Position(line=line, character=start),
                    end=Position(line=line, character=end),
                ),
                new_text=insert_text,
            ),
        )

        self.logger.debug("Item %s", lsp.dump(item))
        return item

    def target_object_to_completion_item(
        self, name: str, display_name: str, obj_type: str
    ) -> CompletionItem:
        """Convert a target object to its CompletionItem representation."""

        key = obj_type

        if ":" in key:
            _, key = key.split(":")

        target_type = COMPLETION_TARGETS.get(key, DEFAULT_TARGET)

        return CompletionItem(
            label=name,
            kind=target_type.kind,
            detail=str(display_name),
            insert_text=target_type.insert_fmt.format(name=name),
        )


def setup(rst: lsp.RstLanguageServer):
    role_completion = Roles(rst)
    rst.add_feature(role_completion)



"""Logic around directive completions goes here."""
import importlib
import inspect
import re
from typing import List
from typing import Tuple
from typing import Union

from docutils.parsers.rst import Directive
from docutils.parsers.rst import directives
from pygls.lsp.types import CompletionItem
from pygls.lsp.types import CompletionItemKind
from pygls.lsp.types import InsertTextFormat
from pygls.lsp.types import Position
from pygls.lsp.types import Range
from pygls.lsp.types import TextEdit
from pygls.workspace import Document

import esbonio.lsp as lsp
from esbonio.lsp.sphinx import get_domains


DIRECTIVE = re.compile(
    r"""
    (?P<indent>\s*)             # directives can be indented
    (?P<directive>\.\.          # start with a comment
    [ ]                         # separated by a space
    (?P<domain>[\w]+:)?         # with an optional domain namespace
    (?P<name>[\w-]+))           # with a name
    ::
    ([\s]+(?P<argument>.*$))?   # some directives may take an argument
    """,
    re.VERBOSE,
)
"""A regular expression that matches a complete, valid directive declaration. Not
including any options or content."""


PARTIAL_DIRECTIVE = re.compile(
    r"""
    (?P<indent>^\s*)          # directives can be indented
    \.\.                      # start with a commment
    (?P<prefix>[ ]?)          # may have a space
    (?P<domain>[\w]+:)?       # with an optional domain namespace
    (?P<name>[\w-]+)?         # with an optional name
    $
    """,
    re.VERBOSE,
)
"""A regular expression that matches a partial directive declaraiton. Used when
generating auto complete suggestions."""


PARTIAL_DIRECTIVE_OPTION = re.compile(
    r"""
    (?P<indent>\s+)   # directive options must only be preceeded by whitespace
    :                 # they start with a ':'
    (?P<name>[\w-]*)  # they have a name
    $
    """,
    re.VERBOSE,
)
"""A regular expression that matches a partial directive option. Used when generating
auto complete suggestions."""


class Directives(lsp.LanguageFeature):
    """Directive support for the language server."""

    def initialize(self, options: lsp.InitializationOptions):
        self.discover()

    def discover(self):
        """Build an index of all available directives and the their options."""
        ignored_directives = ["restructuredtext-test-directive"]

        # Find directives that have been registered directly with docutils.
        found_directives = {**directives._directive_registry, **directives._directives}

        # Find directives under Sphinx domains
        for prefix, domain in get_domains(self.rst.app):
            fmt = "{prefix}:{name}" if prefix else "{name}"

            for name, directive in domain.directives.items():
                key = fmt.format(name=name, prefix=prefix)
                found_directives[key] = directive

        self.directives = {
            k: self.resolve_directive(v)
            for k, v in found_directives.items()
            if k not in ignored_directives
        }

        self.options = {
            k: self.options_to_completion_items(k, v)
            for k, v in self.directives.items()
        }

        self.logger.info("Discovered %s directives", len(self.directives))
        self.logger.debug("Directives: %s", list(self.directives.keys()))

    def resolve_directive(self, directive: Union[Directive, Tuple[str]]):

        # 'Core' docutils directives are returned as tuples (modulename, ClassName)
        # so its up to us to resolve the reference
        if isinstance(directive, tuple):
            mod, cls = directive

            modulename = "docutils.parsers.rst.directives.{}".format(mod)
            module = importlib.import_module(modulename)
            directive = getattr(module, cls)

        return directive

    suggest_triggers = [PARTIAL_DIRECTIVE, PARTIAL_DIRECTIVE_OPTION]
    """Regular expressions that match lines that we want to offer autocomplete
    suggestions for."""

    def suggest(
        self, match: "re.Match", doc: Document, position: Position
    ) -> List[CompletionItem]:
        self.logger.debug("Trigger match: %s", match)
        groups = match.groupdict()

        if "domain" in groups:
            return self.suggest_directives(match, position)

        return self.suggest_options(match, doc, position)

    def suggest_directives(self, match, position) -> List[CompletionItem]:
        self.logger.debug("Suggesting directives")

        domain = match.groupdict()["domain"] or ""
        items = []

        for name, directive in self.directives.items():

            if not name.startswith(domain):
                continue

            item = self.directive_to_completion_item(name, directive, match, position)
            items.append(item)

        return items

    def suggest_options(
        self, match: "re.Match", doc: Document, position: Position
    ) -> List[CompletionItem]:

        groups = match.groupdict()

        self.logger.info("Suggesting options")
        self.logger.debug("Match groups: %s", groups)

        indent = groups["indent"]

        # Search backwards so that we can determine the context for our completion
        linum = position.line - 1
        line = doc.lines[linum]

        while line.startswith(indent):
            linum -= 1
            line = doc.lines[linum]

        # Only offer completions if we're within a directive's option block
        match = DIRECTIVE.match(line)

        self.logger.debug("Context line:  %s", line)
        self.logger.debug("Context match: %s", match)

        if not match:
            return []

        domain = match.group("domain") or ""
        name = f"{domain}{match.group('name')}"

        return self.options.get(name, [])

    def directive_to_completion_item(
        self, name: str, directive: Directive, match: "re.Match", position: Position
    ) -> CompletionItem:
        """Convert an rst directive to its CompletionItem representation.

        Previously, it was fine to pre-convert directives into their completion item
        representation during the :meth:`discover` phase. However a number of factors
        combined to force this to be something we have to compute specifically for each
        completion site.

        It all stems from directives that live under a namespaced domain e.g.
        ``.. c:macro::``. First in order to get trigger character completions for
        directives, we need to allow users to start typing the directive name
        immediately after the second dot and have the CompletionItem insert the leading
        space. Which is exactly what we used to do, setting
        ``insert_text=" directive::"`` and we were done.

        However with domain support, we introduced the possibility of a ``:`` character
        in the name of a directive. You can imagine a scenario where a user types in a
        domain namespace, say ``py:`` in order to filter down the list of options to
        directives that belong to that namespace. With ``:`` being a trigger character
        for role completions and the like, this would cause editors like VSCode to issue
        a new completion request ignoring the old one.

        That isn't necessarily the end of the world, but with CompletionItems assuming
        that they were following the ``..`` characters, the ``insert_text`` was no
        longer correct leading to broken completions like ``..py: py:function::``.

        In order to handle the two scenarios, conceptually the easiest approach is to
        switch to using a ``text_edit`` and replace the entire line with the correct
        text. Unfortunately in practice this was rather fiddly.

        Upon first setting the ``text_edit`` field VSCode suddenly stopped presenting
        any options! After much debugging, head scratching and searching, I eventually
        found a `couple <https://github.com/microsoft/vscode/issues/38982>`_ of
        `issues <https://github.com/microsoft/vscode/issues/41208>`_ that hinted as to
        what was happening.

        I **think** what happens is that since the ``range`` of the text edit extends
        back to the start of the line VSCode considers the entire line to be the filter
        for the CompletionItems so it's looking to select items that start with ``..``
        - which is none of them!

        To work around this, we additionaly need to set the ``filter_text`` field so
        that VSCode computes matches against that instead of the label. Then in order
        for the items to be shown the value of that field needs to be ``..my:directive``
        so it corresponds with what the user has actually written.

        Parameters
        ----------
        name:
           The name of the directive as a user would type in an reStructuredText
           document
        directive:
           The class definition that implements the Directive's behavior
        match:
           The regular expression match object that represents the line we are providing
           the autocomplete suggestions for.
        position:
           The position in the source code where the autocompletion request was sent
           from.
        """
        groups = match.groupdict()
        prefix = groups["prefix"]
        indent = groups["indent"]

        documentation = inspect.getdoc(directive)

        # Ignore directives that do not provide their own documentation.
        if any(
            [
                documentation.startswith("Base class for reStructuredText directives."),
                documentation.startswith("A base class for Sphinx directives."),
            ]
        ):
            documentation = None

        # TODO: Give better names to arguments based on what they represent.
        args = " ".join(
            "${{{0}:arg{0}}}".format(i)
            for i in range(1, directive.required_arguments + 1)
        )

        return CompletionItem(
            label=name,
            kind=CompletionItemKind.Class,
            detail="directive",
            documentation=documentation,
            filter_text=f"..{prefix}{name}",
            insert_text_format=InsertTextFormat.Snippet,
            text_edit=TextEdit(
                range=Range(
                    start=Position(line=position.line, character=0),
                    end=Position(line=position.line, character=position.character - 1),
                ),
                new_text=f"{indent}.. {name}:: {args}",
            ),
        )

    def options_to_completion_items(
        self, name: str, directive: Directive
    ) -> List[CompletionItem]:
        """Convert a directive's options to a list of completion items.

        Unfortunately, the ``autoxxx`` family of directives are a little different.
        Each ``autoxxxx`` directive name resolves to the same ``AutodocDirective`` class.
        That paricular directive does not have any options, instead the options are
        held on the particular Documenter that documents that object type.

        This method does the lookup in order to determine what those options are.

        Parameters
        ----------
        name:
           The name of the directive as it appears in an rst file.
        directive:
           The directive whose options we are creating completions for.
        """

        options = directive.option_spec

        # autoxxx directives require special handlng.
        if name.startswith("auto") and self.rst.app:
            self.logger.debug("Processing options for '%s' directive", name)
            name = name.replace("auto", "")

            self.logger.debug("Documenter name is '%s'", name)
            documenter = self.rst.app.registry.documenters.get(name, None)

            if documenter is not None:
                options = documenter.option_spec

        if options is None:
            return []

        return [
            CompletionItem(
                label=opt,
                detail="option",
                kind=CompletionItemKind.Field,
                insert_text=f"{opt}: ",
            )
            for opt in options
        ]


def setup(rst: lsp.RstLanguageServer):

    directive_completion = Directives(rst)
    rst.add_feature(directive_completion)



import argparse
import sys

import esbonio.lsp as lsp
from esbonio.lsp import __version__


def start_server(args):
    """Start the language server."""

    server = lsp.create_language_server(lsp.BUILTIN_MODULES)

    if args.port:
        server.start_tcp("localhost", args.port)
    else:
        server.start_io()


cli = argparse.ArgumentParser(prog="esbonio", description="The Esbonio language server")

cli.add_argument(
    "-p",
    "--port",
    type=int,
    default=None,
    help="start a TCP instance of the language server listening on the given port ",
)

cli.add_argument(
    "--version", action="store_true", help="print the current version and exit"
)


def main():
    args = cli.parse_args()

    if args.version:
        print("v{}".format(__version__))
        sys.exit(0)

    start_server(args)



import enum
import importlib
import json
import logging
import pathlib
import textwrap
from typing import List
from typing import Optional
from urllib.parse import unquote
from urllib.parse import urlparse

from pydantic import BaseModel
from pydantic import Field
from pygls.lsp.methods import COMPLETION
from pygls.lsp.methods import INITIALIZE
from pygls.lsp.methods import INITIALIZED
from pygls.lsp.methods import TEXT_DOCUMENT_DID_OPEN
from pygls.lsp.methods import TEXT_DOCUMENT_DID_SAVE
from pygls.lsp.types import CompletionList
from pygls.lsp.types import CompletionOptions
from pygls.lsp.types import CompletionParams
from pygls.lsp.types import DidOpenTextDocumentParams
from pygls.lsp.types import DidSaveTextDocumentParams
from pygls.lsp.types import InitializedParams
from pygls.lsp.types import InitializeParams
from pygls.lsp.types import Position
from pygls.server import LanguageServer
from pygls.workspace import Document
from sphinx.application import Sphinx

from esbonio.lsp.logger import LOG_LEVELS
from esbonio.lsp.logger import LogFilter
from esbonio.lsp.logger import LspHandler

__version__ = "0.6.2"


BUILTIN_MODULES = [
    "esbonio.lsp.sphinx",
    "esbonio.lsp.directives",
    "esbonio.lsp.roles",
    "esbonio.lsp.intersphinx",
    "esbonio.lsp.filepaths",
]


class SphinxConfig(BaseModel):
    """Represents both the current Sphinx configuration and also the config options that
    we should create Sphinx with."""

    version: Optional[str]
    """Sphinx's version number."""

    conf_dir: Optional[str] = Field(None, alias="confDir")
    """Can be used to override the default conf.py discovery mechanism."""

    src_dir: Optional[str] = Field(None, alias="srcDir")
    """Can be used to override the default assumption on where the project's rst files are
    located."""

    build_dir: Optional[str] = Field(None, alias="buildDir")
    """Can be used to override the default location for storing build outputs."""

    builder_name: str = Field("html", alias="builderName")
    """The currently used builder."""


class ServerConfig(BaseModel):
    """Configuration options for the server."""

    log_level: Optional[str] = Field("error", alias="logLevel")
    """The logging level for server messages"""

    log_filter: Optional[List[str]] = Field(None, alias="logFilter")
    """A list of logger names to restrict output to."""

    hide_sphinx_output: Optional[bool] = Field(False, alias="hideSphinxOutput")
    """A flag to indicate if Sphinx build output should be omitted from the log."""


class InitializationOptions(BaseModel):
    """The initialization options we can expect to receive from a client."""

    sphinx: Optional[SphinxConfig] = Field(default_factory=SphinxConfig)
    """The ``esbonio.sphinx.*`` namespace of options"""

    server: Optional[ServerConfig] = Field(default_factory=ServerConfig)
    """The ``esbonio.server.*`` namespace of options"""


class LanguageFeature:
    """Base class for language features."""

    def __init__(self, rst: "RstLanguageServer"):
        self.rst = rst
        self.logger = rst.logger.getChild(self.__class__.__name__)


class RstLanguageServer(LanguageServer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.logger = logging.getLogger(__name__)
        """The logger that should be used for all Language Server log entries"""

        self.app: Optional[Sphinx] = None
        """Sphinx application instance configured for the current project."""

        self.on_initialize_hooks = []
        """A list of functions to run on initialization"""

        self.on_initialized_hooks = []
        """A list of functions to run after receiving the initialized notification from the
        client"""

        self.on_save_hooks = []
        """A list of hooks to run on document save."""

        self.completion_handlers = {}
        """The collection of registered completion handlers."""

    def add_feature(self, feature):
        """Add a new feature to the language server."""

        if hasattr(feature, "initialize"):
            self.on_initialize_hooks.append(feature.initialize)

        if hasattr(feature, "initialized"):
            self.on_initialized_hooks.append(feature.initialized)

        if hasattr(feature, "save"):
            self.on_save_hooks.append(feature.save)

        if hasattr(feature, "suggest") and hasattr(feature, "suggest_triggers"):

            for trigger in feature.suggest_triggers:
                handler = feature.suggest

                if trigger in self.completion_handlers:
                    self.completion_handlers[trigger].append(handler)
                else:
                    self.completion_handlers[trigger] = [handler]

    def load_module(self, mod: str):
        # TODO: Handle failures.
        module = importlib.import_module(mod)

        if not hasattr(module, "setup"):
            raise TypeError("Module '{}' missing setup function".format(mod))

        module.setup(self)

    def run_hooks(self, kind: str, *args):
        """Run each hook registered of the given kind."""
        hooks = getattr(self, f"on_{kind}_hooks")

        for hook in hooks:
            self.logger.debug("Running '%s' hook %s", kind, hook)
            hook(*args)

    def _configure_logging(self, config: ServerConfig):
        level = LOG_LEVELS[config.log_level]

        lsp_logger = logging.getLogger("esbonio.lsp")
        lsp_logger.setLevel(level)

        lsp_handler = LspHandler(self)
        lsp_handler.setLevel(level)

        if config.log_filter is not None and len(config.log_filter) > 0:
            lsp_handler.addFilter(LogFilter(config.log_filter))

        formatter = logging.Formatter("[%(name)s] %(message)s")
        lsp_handler.setFormatter(formatter)
        lsp_logger.addHandler(lsp_handler)

        if not config.hide_sphinx_output:
            sphinx_logger = logging.getLogger("esbonio.sphinx")
            sphinx_logger.setLevel(logging.INFO)

            sphinx_handler = LspHandler(self)
            sphinx_handler.setLevel(logging.INFO)

            formatter = logging.Formatter("%(message)s")
            sphinx_handler.setFormatter(formatter)
            sphinx_logger.addHandler(sphinx_handler)


def get_line_til_position(doc: Document, position: Position) -> str:
    """Return the line up until the position of the cursor."""

    try:
        line = doc.lines[position.line]
    except IndexError:
        return ""

    return line[: position.character]


def filepath_from_uri(uri: str) -> pathlib.Path:
    """Given a uri, return the filepath component."""

    uri = urlparse(uri)
    return pathlib.Path(unquote(uri.path))


def dump(obj) -> str:
    """Debug helper function that converts an object to JSON."""

    def default(o):
        if isinstance(o, enum.Enum):
            return o.value

        fields = {}
        for k, v in o.__dict__.items():

            if v is None:
                continue

            # Truncate long strings - but not uris!
            if isinstance(v, str) and not k.lower().endswith("uri"):
                v = textwrap.shorten(v, width=25)

            fields[k] = v

        return fields

    return json.dumps(obj, default=default)


def create_language_server(modules: List[str]) -> RstLanguageServer:
    """Create a new language server instance.

    Parameters
    ----------
    modules:
        The list of modules that should be loaded.
    """
    server = RstLanguageServer()

    for mod in modules:
        server.load_module(mod)

    @server.feature(INITIALIZE)
    def on_initialize(rst: RstLanguageServer, params: InitializeParams):
        options = InitializationOptions(**params.initialization_options)

        # Let there be light...
        rst._configure_logging(options.server)
        rst.logger.info("Language server started.")

        rst.logger.debug("%s: %s", INITIALIZE, dump(params))
        rst.run_hooks("initialize", options)

    @server.feature(INITIALIZED)
    def on_initialized(rst: RstLanguageServer, params: InitializedParams):
        rst.logger.debug("%s: %s", INITIALIZED, dump(params))
        rst.run_hooks("initialized")

    @server.feature(
        COMPLETION, CompletionOptions(trigger_characters=[".", ":", "`", "<", "/"])
    )
    def on_completion(rst: RstLanguageServer, params: CompletionParams):
        """Suggest completions based on the current context."""
        rst.logger.debug("%s: %s", COMPLETION, dump(params))

        uri = params.text_document.uri
        pos = params.position

        doc = rst.workspace.get_document(uri)
        line = get_line_til_position(doc, pos)

        items = []

        for pattern, handlers in rst.completion_handlers.items():
            match = pattern.match(line)
            if match:
                for handler in handlers:
                    items += handler(match, doc, pos)

        return CompletionList(is_incomplete=False, items=items)

    @server.feature(TEXT_DOCUMENT_DID_OPEN)
    def on_open(rst: RstLanguageServer, params: DidOpenTextDocumentParams):
        rst.logger.debug("%s: %s", TEXT_DOCUMENT_DID_OPEN, dump(params))

    @server.feature(TEXT_DOCUMENT_DID_SAVE)
    def on_save(rst: RstLanguageServer, params: DidSaveTextDocumentParams):
        rst.logger.debug("%s: %s", TEXT_DOCUMENT_DID_SAVE, dump(params))
        rst.run_hooks("save", params)

    @server.feature("$/setTrace")
    def on_set_trace(rst: RstLanguageServer, params):
        """Dummy implementation, stops JsonRpcMethodNotFound exceptions."""
        rst.logger.debug("$/setTrace: %s", dump(params))

    @server.feature("$/setTraceNotification")
    def vscode_set_trace(rst: RstLanguageServer, params):
        """Dummy implementation, stops JsonRpcMethodNotFound exceptions."""
        rst.logger.debug("$/setTraceNotification: %s", dump(params))

    return server



"""Completion handler for filepaths.

While :mod:`esbonio.lsp.roles` and :mod:`esbonio.lsp.directives` provide generic
completion handlers for roles and directives, similar to :mod:`esbonio.lsp.intersphinx`
this is a "specialised" module dedicated to providing completion suggestions just for
the roles or directives that accept filepaths as arguments.
"""
import pathlib
import re
from typing import Dict
from typing import List

from pygls.lsp.types import CompletionItem
from pygls.lsp.types import CompletionItemKind
from pygls.lsp.types import Position
from pygls.workspace import Document

import esbonio.lsp as lsp
from esbonio.lsp import LanguageFeature
from esbonio.lsp import RstLanguageServer
from esbonio.lsp.directives import DIRECTIVE
from esbonio.lsp.roles import PARTIAL_ALIASED_TARGET
from esbonio.lsp.roles import PARTIAL_PLAIN_TARGET


# TODO: Would it be better to make the role and directive language features extensible
#       and have this and the intersphinx feature contribute suggestions when they know
#       something?..
class FilepathCompletions(LanguageFeature):
    """Filepath completion support."""

    suggest_triggers = [DIRECTIVE, PARTIAL_PLAIN_TARGET, PARTIAL_ALIASED_TARGET]

    def suggest(
        self, match: "re.Match", doc: Document, position: Position
    ) -> List[CompletionItem]:

        groups = match.groupdict()

        if not self.should_suggest(groups):
            return []

        if "target" in groups:
            partial_path = groups["target"]
        else:
            partial_path = groups["argument"]

        if partial_path.startswith("/"):
            # Absolute paths are relative to the top level source dir.
            candidate_dir = pathlib.Path(self.rst.app.srcdir)

            # Be sure to take off the leading '/' character, otherwise the partial
            # path will wipe out the srcdir part when concatenated..
            partial_path = partial_path[1:]
        else:
            # Otherwise they're relative to the current file.
            filepath = lsp.filepath_from_uri(doc.uri)
            candidate_dir = pathlib.Path(filepath).parent

        candidate_dir /= pathlib.Path(partial_path)

        if partial_path and not partial_path.endswith("/"):
            candidate_dir = candidate_dir.parent

        return [self.path_to_completion_item(p) for p in candidate_dir.glob("*")]

    def should_suggest(self, groups: Dict[str, str]) -> bool:
        """Determines if we should make any suggestions."""

        roles = {"download"}
        directives = {"image", "figure", "include", "literalinclude"}

        return any(
            [
                groups["name"] in roles and "role" in groups,
                groups["name"] in directives and "directive" in groups,
            ]
        )

    def path_to_completion_item(self, path: pathlib.Path) -> CompletionItem:

        kind = CompletionItemKind.Folder if path.is_dir() else CompletionItemKind.File
        return CompletionItem(
            label=str(path.name),
            kind=kind,
            insert_text=f"{path.name}",
        )


def setup(rst: RstLanguageServer):
    filepaths = FilepathCompletions(rst)
    rst.add_feature(filepaths)



"""Code for managing sphinx applications."""
import collections
import hashlib
import logging
import pathlib
import re
import traceback
from typing import Iterator
from typing import Optional
from typing import Tuple
from urllib.parse import quote

import appdirs
from pygls.lsp.types import Diagnostic
from pygls.lsp.types import DiagnosticSeverity
from pygls.lsp.types import DidSaveTextDocumentParams
from pygls.lsp.types import MessageType
from pygls.lsp.types import Position
from pygls.lsp.types import Range
from sphinx import __version__ as __sphinx_version__
from sphinx.application import Sphinx
from sphinx.domains import Domain
from sphinx.util import console

import esbonio.lsp as lsp


PATH_VAR_PATTERN = re.compile(r"^\${(\w+)}/?.*")


PROBLEM_PATTERN = re.compile(
    r"""
    (?P<file>(.*:\\)?[^:]*):   # Capture the path to the file containing the problem
    ((?P<line>\d+):)?          # Some errors may specify a line number.
    \s(?P<type>[^:]*):         # Capture the type of error
    (\s+)?(?P<message>.*)      # Capture the error message
    """,
    re.VERBOSE,
)
"""Regular Expression used to identify warnings/errors in Sphinx's output.

For example::

   /path/to/file.rst: WARNING: document isn't included in any toctree
   /path/to/file.rst:4: WARNING: toctree contains reference to nonexisting document 'changelog',

"""


PROBLEM_SEVERITY = {
    "WARNING": DiagnosticSeverity.Warning,
    "ERROR": DiagnosticSeverity.Error,
}


def get_domains(app: Sphinx) -> Iterator[Tuple[str, Domain]]:
    """Get all the domains registered with an applications.

    Returns a generator that iterates through all of an application's domains,
    taking into account configuration variables such as ``primary_domain``.
    Yielded values will be a tuple of the form ``(prefix, domain)`` where

    - ``prefix`` is the namespace that should be used when referencing items
      in the domain
    - ``domain`` is the domain object itself.
    """

    if app is None:
        return []

    domains = app.env.domains
    primary_domain = app.config.primary_domain

    for name, domain in domains.items():
        prefix = name

        # Items from the standard and primary domains don't require the namespace prefix
        if name == "std" or name == primary_domain:
            prefix = ""

        yield prefix, domain


def expand_conf_dir(root_dir: str, conf_dir: str) -> str:
    """Expand the user provided conf_dir into a real path.

    Here is where we handle "variables" that can be included in the path, currently
    we support

    - ``${workspaceRoot}`` which expands to the workspace root as provided by the
      language client.

    Parameters
    ----------
    root_dir:
       The workspace root path
    conf_dir:
       The user provided path
    """

    match = PATH_VAR_PATTERN.match(conf_dir)
    if not match or match.group(1) != "workspaceRoot":
        return conf_dir

    conf = pathlib.Path(conf_dir).parts[1:]
    return pathlib.Path(root_dir, *conf).resolve()


def get_src_dir(
    root_uri: str, conf_dir: pathlib.Path, config: lsp.SphinxConfig
) -> pathlib.Path:
    """Get the src dir to use based on the given config.

    By default the src dir will be the same as the conf dir, but this can
    be overriden in the given config.

    There are a number of "variables" that can be included in the path, currently
    we support

    - ``${workspaceRoot}`` which expands to the workspace root as provided by the
      language client.
    - ``${confDir}`` which expands to the configured config dir.

    Parameters
    ----------
    root_uri:
       The workspace root uri
    conf_dir:
       The project's conf dir
    config:
       The user's configuration.
    """

    if not config.src_dir:
        return conf_dir

    src_dir = config.src_dir
    root_dir = lsp.filepath_from_uri(root_uri)

    match = PATH_VAR_PATTERN.match(src_dir)
    if match and match.group(1) == "workspaceRoot":
        src = pathlib.Path(src_dir).parts[1:]
        return pathlib.Path(root_dir, *src).resolve()

    if match and match.group(1) == "confDir":
        src = pathlib.Path(src_dir).parts[1:]
        return pathlib.Path(conf_dir, *src).resolve()

    return src_dir


def get_build_dir(conf_dir: pathlib.Path, config: lsp.SphinxConfig) -> pathlib.Path:

    if config.build_dir is None:
        # Try to pick a sensible dir based on the project's location
        cache = appdirs.user_cache_dir("esbonio", "swyddfa")
        project = hashlib.md5(str(conf_dir).encode()).hexdigest()

        return pathlib.Path(cache) / project

    build_dir = pathlib.Path(config.build_dir)

    # Strangely for windows paths, there's an extra leading slash which we have to
    # remove ourselves.
    if isinstance(build_dir, pathlib.WindowsPath) and str(build_dir).startswith("\\"):
        build_dir = pathlib.Path(str(build_dir)[1:])

    return build_dir


def find_conf_dir(root_uri: str, config: lsp.SphinxConfig) -> Optional[pathlib.Path]:
    """Attempt to find Sphinx's configuration file in the given workspace."""

    root = lsp.filepath_from_uri(root_uri)

    if config.conf_dir:
        return expand_conf_dir(root, config.conf_dir)

    # Strangely for windows paths, there's an extra leading slash which we have to
    # remove ourselves.
    if isinstance(root, pathlib.WindowsPath) and str(root).startswith("\\"):
        root = pathlib.Path(str(root)[1:])

    # Try and find Sphinx's conf.py file
    ignore_paths = [".tox", "site-packages"]

    for candidate in root.glob("**/conf.py"):

        # Skip files that obviously aren't part of the project
        if any(path in str(candidate) for path in ignore_paths):
            continue

        return candidate.parent


class DiagnosticList(collections.UserList):
    """A list type dedicated to holding diagnostics.

    This is mainly to ensure that only one instance of a diagnostic ever gets
    reported.
    """

    def append(self, item: Diagnostic):

        if not isinstance(item, Diagnostic):
            raise TypeError("Expected Diagnostic")

        for existing in self.data:
            fields = [
                existing.range == item.range,
                existing.message == item.message,
                existing.severity == item.severity,
                existing.code == item.code,
                existing.source == item.source,
            ]

            if all(fields):
                # Item already added, nothing to do.
                return

        self.data.append(item)


class SphinxManagement(lsp.LanguageFeature):
    """A LSP Server feature that manages the Sphinx application instance for the
    project."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.diagnostics = {}
        """A place to keep track of diagnostics we can publish to the client."""

        self.config: Optional[lsp.SphinxConfig] = None
        """The client's ``esbonio.sphinx.*`` configuration."""

        self._conf_dir = None
        """The directory containing Sphinx's ``conf.py``.

        The source of truth for this value should be the Sphinx application itself,
        **you should not depend on this value**.

        The only use case for this field is when a user's config is broken and there
        currently isn't a valid application object, we use this field to determine when
        a user has edited their ``conf.py`` and we should try to restart the server.
        """

        self.sphinx_log = logging.getLogger("esbonio.sphinx")
        """The logger that should be used by a Sphinx application"""

    def initialize(self, options: lsp.InitializationOptions):

        self.config = options.sphinx
        self.logger.debug("SphinxConfig %s", self.config.dict())
        self.create_app(self.config)
        self.build_app()

    def initialized(self):

        if not self.rst.app:
            return

        app = self.rst.app
        params = lsp.SphinxConfig(
            version=__sphinx_version__,
            confDir=app.confdir,
            srcDir=app.srcdir,
            buildDir=app.outdir,
            builderName=app.builder.name,
        )
        self.rst.logger.debug("Final Sphinx Config: %s", params.dict(by_alias=True))
        self.rst.send_notification("esbonio/sphinxConfiguration", params)

    def save(self, params: DidSaveTextDocumentParams):

        filepath = lsp.filepath_from_uri(params.text_document.uri)

        # There may not be an application instance - the user's config could
        # be broken...
        if not self.rst.app:

            # ...did thry try to fix it?
            if self._conf_dir and filepath == (self._conf_dir / "conf.py"):
                self.create_app(self.config)

        # The user has updated their conf.py, we need to recreate the application.
        elif filepath == pathlib.Path(self.rst.app.confdir) / "conf.py":
            self.create_app(self.config)

        else:
            self.reset_diagnostics(str(filepath))

        self.build_app()

    def create_app(self, config: lsp.SphinxConfig):
        """Initialize a Sphinx application instance for the current workspace."""
        self.rst.logger.debug("Workspace root %s", self.rst.workspace.root_uri)

        self.diagnostics = {}
        conf_dir = find_conf_dir(self.rst.workspace.root_uri, config)

        if conf_dir is None:
            self.rst.show_message(
                'Unable to find your project\'s "conf.py", features that depend on '
                + "Sphinx will be unavailable",
                msg_type=MessageType.Warning,
            )
            return

        builder_name = config.builder_name
        src_dir = get_src_dir(self.rst.workspace.root_uri, conf_dir, config)
        build_dir = get_build_dir(conf_dir, config)
        doctree_dir = pathlib.Path(build_dir) / "doctrees"
        build_dir /= builder_name

        self.rst.logger.debug("Config dir %s", conf_dir)
        self.rst.logger.debug("Src dir %s", src_dir)
        self.rst.logger.debug("Build dir %s", build_dir)
        self.rst.logger.debug("Doctree dir %s", doctree_dir)

        # Disable color escape codes in Sphinx's log messages
        console.nocolor()

        try:
            self.rst.app = Sphinx(
                srcdir=src_dir,
                confdir=conf_dir,
                outdir=build_dir,
                doctreedir=doctree_dir,
                buildername="html",
                status=self,
                warning=self,
            )
        except Exception:
            self._conf_dir = conf_dir

            self.sphinx_log.error(traceback.format_exc())
            self.rst.show_message(
                message="Unable to initialize Sphinx, see output window for details.",
                msg_type=MessageType.Error,
            )

    def build_app(self):

        if not self.rst.app:
            return

        try:
            self.rst.app.build()
        except Exception:
            message = "Unable to build documentation, see output window for details."

            self.sphinx_log.error(traceback.format_exc())
            self.rst.show_message(
                message=message,
                msg_type=MessageType.Error,
            )

        self.rst.send_notification("esbonio/buildComplete", {})
        self.report_diagnostics()

    def report_diagnostics(self):
        """Publish the current set of diagnostics to the client."""

        for doc, diagnostics in self.diagnostics.items():

            if not doc.startswith("/"):
                doc = "/" + doc.replace("\\", "/")

            uri = f"file://{quote(doc)}"
            self.logger.debug("Publishing diagnostics for document: %s", uri)
            self.rst.publish_diagnostics(uri, diagnostics.data)

    def reset_diagnostics(self, filepath: str):
        """Reset the list of diagnostics for the given file.

        Parameters
        ----------
        filepath:
           The filepath that the diagnostics should be reset for.
        """
        self.diagnostics[filepath] = DiagnosticList()

    def write(self, line):
        """This method lets us catch output from Sphinx."""

        match = PROBLEM_PATTERN.match(line)
        if match:
            filepath = match.group("file")
            severity = PROBLEM_SEVERITY.get(
                match.group("type"), PROBLEM_SEVERITY["ERROR"]
            )
            diagnostics = self.diagnostics.get(filepath, None)

            if diagnostics is None:
                diagnostics = DiagnosticList()

            try:
                line_number = int(match.group("line"))
            except (TypeError, ValueError) as exc:
                self.logger.debug(
                    "Unable to parse line number: '%s' - %s", match.group("line"), exc
                )

                line_number = 1

            range_ = Range(
                start=Position(line=line_number - 1, character=0),
                end=Position(line=line_number, character=0),
            )
            diagnostics.append(
                Diagnostic(
                    range=range_,
                    message=match.group("message"),
                    severity=severity,
                    source="sphinx",
                )
            )

            self.diagnostics[filepath] = diagnostics

        self.sphinx_log.info(line)


def setup(rst: lsp.RstLanguageServer):
    sphinx_management = SphinxManagement(rst)
    rst.add_feature(sphinx_management)
